// Formatted streams.
// Norman Gray <norman@astro.gla.ac.uk>
//
// Produces output, and reads input, in one of the formats listed in
// FIOstream::Formats.
//
// Reasonably thorough, but rather primitive, error checking,
// mostly in the form of assertions which dump core when they fail.

#ifndef NULL
#define NULL 0			// gcc include files seem to have some errors
#endif

#include "FIOstream.h"
#include "btoe.h"
#include <ctype.h>

FIOstream::Formats FIOstream::ioType_ = FIOstream::Binary;
const char *FIOstream::base64Coding = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
char *FIOstream::base64DeCoding = 0;

FIOstream::~FIOstream()
{
  if (isOpen_)
  {
    close();
    isOpen_ = false;
  }
  moreInput_ = false;
}

// Will _not_ reopen a file if it's already open
bool FIOstream::open (string name, Mode RW)
{
  if (!isOpen_)
  {
    isWrite_ = (RW == Write);
    F_ = fopen (name.c_str(), (isWrite_ ? "w" : "r"));
    isOpen_ = (F_ != NULL);
    moreInput_ = isOpen_;
  }
  return isOpen_;
}

void FIOstream::close()
{
  if (isOpen_)
  {
    if (isWrite_)
    {
      if (ioType_ == Base64)
	flushBase64();
      else if (ioType_ == Words)
	flushWords();
      fflush (F_);		// largely unnecessary, but...
    }
    fclose(F_);
    F_ = NULL;
    isOpen_ = false;
    // don't fiddle with moreInput_ here!
  }
}

bool FIOstream::eof() const
{
  // readers below immediately close file when they find EOF
  return !moreInput_;
}

void FIOstream::write (unsigned char ch)
{
  assert (isWrite_);		// wrong: better would be to throw an exception

  if (isOpen_)
    if (ioType_ == Binary)
      fputc (ch, F_);
    else if (ioType_ == Base64)
      toBase64(ch);
    else if (ioType_ == Words)
      toWords(ch);
}

unsigned char FIOstream::read ()
{
  assert (!isWrite_);		// wrong: better would be to throw an exception

  unsigned char t;
  if (!eof())
  {
    if (ioType_ == Binary)
    {
      int ti = fgetc (F_);
      if (ti == EOF)
      {
	close ();
	moreInput_ = false;
	t = 0;			// always succeed
      }
      else
	t = static_cast<unsigned char>(ti);
    }
    else if (ioType_ == Base64)
      t = fromBase64();
    else
    {
      assert (ioType_ == Words);
      t = fromWords();
    }
  }
  else
    t = 0;

  return t;
}

// See RFC 2045, section 6.8
void FIOstream::toBase64 (unsigned char ch)
{
  assert (nochars_ < 3);
  assert (isWrite_);

  ochars_[nochars_] = ch;
  nochars_++;
  if (nochars_ == 3)		// write them
  {
    unsigned char t;
    fputc (base64Coding[(ochars_[0] & 0xfc) >> 2], F_);
    t = ((ochars_[0] & 0x03) << 4) | ((ochars_[1] & 0xf0) >> 4);
    fputc (base64Coding[t], F_);
    t = ((ochars_[1] & 0x0f) << 2) | ((ochars_[2] & 0xc0) >> 6);
    fputc (base64Coding[t], F_);
    fputc (base64Coding[ochars_[2] & 0x3f], F_);

    nochars_ = 0;

    nquartets_++;
    if (nquartets_ >= 19)
    {
      fputs ("\r\n", F_);
      nquartets_ = 0;
    }
  }
}

void FIOstream::flushBase64 ()
{
  assert (isWrite_);

  for (unsigned int i=nochars_; i<3; i++)
    ochars_[i] = 0;		// fill with zero bytes

  if (nochars_ == 0)
  {
    fputs ("\r\n", F_);
    return;
  }

  if (nochars_ > 0)
  {
    unsigned char t;
    fputc (base64Coding[(ochars_[0] & 0xfc) >> 2], F_);
    t = ((ochars_[0] & 0x03) << 4) | ((ochars_[1] & 0xf0) >> 4);
    fputc (base64Coding[t], F_);
  }

  if (nochars_ > 1)
  {
    unsigned char t;
    t = ((ochars_[1] & 0x0f) << 2) | ((ochars_[2] & 0xc0) >> 6);
    fputc (base64Coding[t], F_);
  }
  else
    fputc ('=', F_);

  fputs ("=\r\n", F_);

  nochars_ = 0;
}

// Decode base64.  Little error checking, since we assume we wrote the file 
// correctly!
unsigned char FIOstream::fromBase64 (void)
{
  assert (!isWrite_);
  if (! moreInput_)
    return 0;

  // nochars_ is the number of converted but unread characters in the buffer
  assert (nochars_ < 3);
  if (nochars_ == 0)		// read some more
  {
    if (! isOpen_)		// end of file and end of buffer: all gone!
    {
      moreInput_ = false;
      return 0;
    }
    unsigned char c[4];
    int nequals = 0;
    for (unsigned int i=0; i<4; i++)
      c[i] = getBase64Character();
    if (!isOpen_)
      return 0;
    for (unsigned int i=0; i<4; i++)
    {
      if (c[i] == '=')		// EOD pad character
	nequals++;		// '=' maps to \0, so nothing else to do
      c[i] = base64DeCoding[c[i]];
    }
    ochars_[0] = ((c[0] << 2) | ((c[1] & 0x30) >> 4));
    ochars_[1] = (((c[1] & 0x0f) << 4) | ((c[2] & 0x3c) >> 2));
    ochars_[2] = (((c[2] & 0x03) << 6) | c[3]);

    if (nequals == 0)		// normal case
      nochars_ = 3;
    else			// we found the pad characters at end of file
    {
      // discard pad characters, and shuffle so that there are only
      // (3-nequals) characters left to read
      if (nequals == 2)
      {
	ochars_[2] = ochars_[0];
	nochars_ = 1;
      }
      else
      {
	assert (nequals == 1);
	ochars_[2] = ochars_[1];
	ochars_[1] = ochars_[0];
	nochars_ = 2;
      }
    }
  }
  return ochars_[3-(nochars_--)];
}

// Get a base64 character from the input, skipping all non-base64
// characters, and also skipping any `#...' comments.
unsigned char FIOstream::getBase64Character()
{
  if (! isOpen_)
    return 0;			// always succeed

  if (base64DeCoding == 0)	// initialise it
  {
    base64DeCoding = new char[256];
    for (unsigned int i=0; i<256; i++) // set everything to charIgnore
      base64DeCoding[i] = charIgnore;
    for (unsigned int i=0; i<64; i++) // ... except legal characters
      base64DeCoding[base64Coding[i]] = i;
    base64DeCoding[static_cast<unsigned int>('=')] = 0;
			// ... the pad character (actual map value irrelevant)
    base64DeCoding[static_cast<unsigned int>('#')] = charComment;
    base64DeCoding[static_cast<unsigned int>('\r')] = charEOL;
    base64DeCoding[static_cast<unsigned int>('\n')] = charEOL;
			// ... and special characters
  }

  int ic;
  do
  {
    ic = fgetc (F_);
    if (base64DeCoding[ic] == charComment)
      do			// read to end-of-line
	ic = fgetc (F_);
      while (ic != EOF && base64DeCoding[ic] != charEOL);

    if (ic == EOF)
    {
      close ();
      return 0;			// JUMP OUT
    }
  }
  while (base64DeCoding[ic] > 63);
  return static_cast<unsigned char>(ic);
}
  
void FIOstream::toWords (unsigned char ch)
{
  assert (isWrite_);

  ochars_[nochars_] = ch;
  nochars_++;
  if (nochars_ == 8)		// write them
  {
    btoe (ebuf_, ochars_);
    fputs (ebuf_, F_);
    fputc ('\n', F_);
    nochars_ = 0;
  }
}

void FIOstream::flushWords ()
{
  assert (isWrite_);
  if (nochars_ == 0)
    return;

  for (unsigned int i=nochars_; i<8; i++)
    ochars_[i] = 0;		// fill with zero bytes -- OK for string input
  btoe (ebuf_, ochars_);
  fputs (ebuf_, F_);
  fputc ('\n', F_);
  nochars_ = 0;
}

unsigned char FIOstream::fromWords ()
{
  assert (!isWrite_);

  if (! moreInput_)
    return 0;			// always succeed

  // nochars_ is the number of converted but unread characters in the buffer
  assert (nochars_ < 8);
  if (nochars_ == 0)		// read some more
  {
    if (!isOpen_)		// ... but no more available
    {
      moreInput_ = false;
      return 0;			// always succeed
    }

    char words[6*5+2];
    int buflen = sizeof(words);
    int wi = 0;
    for (int nowords=0; nowords < 6; nowords++)
    {
      int ic;
      do
	ic = fgetc (F_);
      while (isspace(ic));
      if (ic == EOF)
      {
	assert (nowords == 0);	// should only happen at beginning of block
	close();
	return 0;		// LEAP OUT
      }
      do
      {
	words[wi++] = static_cast<char>(ic);
	assert (wi < buflen);	// 6 well-formed words (<= 4 chars) should fit
	ic = fgetc (F_);
      }
      while (isalpha(ic));
      if (ic == EOF)
      {
	assert (nowords == 5);	// should be only when we've found 6 words
	close();
      }
      words[wi++] = ' ';
    }
    words[wi-1] = '\0';

    int etob_ret = etob (ochars_, words);
    if (etob_ret != 1)
    {
      cerr << "Error converting words '" << words << "'\n";
      return 0;
    }
    nochars_ = 8;
  }
  return ochars_[8-(nochars_--)];
}
