#ifndef _SecretShare_h
#define _SecretShare_h

#include "RNGenerator.h"
#include "Polynomial.h"
#include "GF256.h"
#include <string>
#include <iostream.h>
#include "Point.h"
#include "FIOstream.h"

using namespace std;		// for fussy compilers

// It's desirable for this secret sharing algorithm to have the following
// attributes:
// 
// 1) Entropy attributable to a purely random data stream.  This allows for
//    the keys to be embedded using steganographic methods, and even if those
//    methods are known, the content is indistinguishable from other encoded
//    random data stream (Stealth PGP, for example).
// 2) Unpredictable key sizes.  Thus, even given a known steganographic 
//    extraction technique, it won't be clear which keys go together, or, 
//    for that matter, how many keys are needed to extract the content.
 
// Solution for (1):

// To encode a k-bit block, the modulus m must be prime and >= 2**k.  
// However, m will always be larger than 2**k (because it's even).  Thus,
// there's a chance that f(x) (mod m) will overflow a k-bit block (by
// being >= 2**k).  The best solution is to use a large k, and the smallest
// acceptable m for that k.

// For example, 1024-bit blocks, with the smallest prime modulus > 
// 2^1024, would provide a very small chance of overflowing the modulus,
// (approximately 10**-306) and this could be corrected for by repicking 
// any x whose f(x) overflows the modulus.  This will not provide insight 
// into the nature of the generating polynomial, as it's never clear 
// whether a particular x was skipped in key generation, or simply not 
// chosen.
//

// Table of primes used: Primality from Mathematica's PrimeQ[]
// 
// Bytes  Number   Smallest Prime 
//===============================
// 1      2^8      2^8+1
// 2      2^16     2^16+1
// 4      2^32     2^32+15
// 8      2^64     2^64+13
// 16     2^128    2^128+51
// 32     2^256    2^256+297
// 64     2^512    2^512+75
// 128    2^1024   2^1024+643
// 256    2^2048   2^2048+981

// Hal Finney (hal@rain.org) suggested solving this problem using a Galois
// field, GF(256).  (Also called a finite field).  A very good idea!  So
// I implemented a fast GF256 using lookup tables.


class SecretShare {
public:
  SecretShare(int keys, int keysNeeded, bool stego = false);
  ~SecretShare();

  void encode(istream& stream);
  void encode(unsigned char ch);
  void createKeys(const Polynomial<GF256>& p);

  //bool decode(ostream& out, ifstream* inputs);
  bool decode(ostream& out, FIOstream** inputs);
  bool decode(ostream& out, const Point<GF256>* inputs);

  void setPrefix(const string& newPrefix);
  string getPrefix() const;

  void setRNG(const RNGenerator& rn);

private:
  int bytes;
  int keys;
  int keysNeeded;
  bool stego;

  // This value must be even, because we push characters two at a time, and
  // don't check overflow before this.
  static const int bufferSize = 4096;


  char** buffer;
  int* charsPushed;

  RNGenerator rand;
  string prefix;

  typedef FIOstream* FIOstreamp;
  //FIOstream **ofiles;
  FIOstreamp *ofiles;

  void createBuffers();
  void flush();
  void deleteBuffers();
 
  //void appendNoise(ofstream& out);
  void appendNoise(FIOstream*);

  //void outfile(ofstream& file, int i, bool append =false);
  //void infile(ifstream& file, int i);

  bool keyUsed(int* usedKeys, int index, int x) const;
};

#endif // _SecretShare_h


