#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <time.h>

#include <string>
#include <fstream.h>
#include <strstream.h>

#include "RNGenerator.h"
#include "SecretShare.h"
#include "FIOstream.h"
#include "version.h"

static void decode(int argc, char** argv, const string& output, int k);
static void encode(char* name, bool isFilename,
		   const string& prefix, int N, 
		   int k, bool isExecutable, const string& randomSource,
		   bool useStego);
static void setInt(const char* str, int& val);
static void usage();

main(int argc, char** argv)
{
  //string output("decoded");
  string output("");
  string prefix("key.");
  string randomSource("/dev/random");
  int ich;
  char ch;
  bool doDecode = false;
  bool isExecutable = false;
  bool isFilename = true;
  bool useStego = false;
  int k = -1;
  int N = -1;

  FIOstream::setFormat (FIOstream::Binary);	// default

  while ((ich = getopt(argc, argv, "b:N:k:o:p:r:R:t:dsSV")) != EOF) {
    ch = static_cast<char>(ich);
    // ich must be int for compilers where `char' is unsigned
    switch (ch) {

    case 'N':
      setInt(optarg, N);
      break;

    case 'k':
      setInt(optarg, k);
      break;

    case 'd':
      doDecode = true;
      break;

    case 'o':
      output = optarg;
      break;

    case 'p':
      prefix = optarg;
      break;

    case 'r':
      isExecutable = false;
      randomSource = optarg;
      break;

    case 'R':
      isExecutable = true;
      randomSource = optarg;
      break;

     case 's':
      useStego = true;
      break;

     case 'S':
      isFilename = false;
      break;

     case 't':
      if (strcmp (optarg, "binary") == 0)
	FIOstream::setFormat (FIOstream::Binary);
      else if (strcmp (optarg, "base64") == 0)
	FIOstream::setFormat (FIOstream::Base64);
      else if (strcmp (optarg, "words") == 0)
	FIOstream::setFormat (FIOstream::Words);
      else
	usage();
      break;

     case 'V':
      cout << version_string << endl;
      exit (0);
      break;

    default:
      usage();
      break;
    }
  }

  argc -= optind;
  argv += optind;

  if ((N > 0) && (k < 0))
    k = N;

  if ((N < 0) && (k > 0))
    N = k;

  if (doDecode) {
    if (k < 0) 
      k = argc;
    
    decode(argc, argv, output, k);
  } else {
    if (!argc)
      usage();
    encode(argv[0], isFilename, prefix, N, k, isExecutable, randomSource, useStego);
  }
}

static void encode(char* name, bool isFilename,
		   const string& prefix, 
		   int N, int k, bool isExecutable, const string& randomSource,
		   bool useStego)
{
  SecretShare ss(N, k, useStego);

  if (k > N) {
    cout << "Error: (N > k) -- Not enough keys to reconstruct secret." << endl;
    exit(0);
  }

  if (k < 1) {
    cout << "Error: k has bad value (" << k << ")." << endl;
    exit(0);
  } 

  if (k == 1) {
    cout << "Warning: Every key will encode the secret literally." << endl;
  }

  ss.setPrefix(prefix);

  RNGenerator rand;

  if (isExecutable) 
    rand.setSource(RNGenerator::Executable, randomSource);
  else
    rand.setSource(RNGenerator::File, randomSource);

  ss.setRNG(rand);

  if (isFilename)
  {
    ifstream input(name);
    if (!input) {
      cout << "Error: Cannot open '" << name << "'" << endl;
      return;
    }

    ss.encode(input);
  }
  else
  {
    istrstream input(name);
    ss.encode(input);
  }

}

static void decode(int nfiles, char** filenames, const string& output, int k)
{
  SecretShare ss(k, k);

  //ifstream* files = new ifstream[nfiles];
  FIOstream** files = new (FIOstream*)[nfiles];
  ofstream outstream;
  ostream *out;
  // if output string present, redirect output to a file, otherwise cout
  if (output.length() == 0)
    out = dynamic_cast<ostream*>(&cout);
  else
  {
    outstream.open(output.c_str());
    out = dynamic_cast<ostream*>(&outstream);
  }
 
  for (int i = 0; i < nfiles; i++) {
    //files[i].open(filenames[i]);
    files[i] = new FIOstream (filenames[i], FIOstream::Read);
    if (!files[i]) {
      cout << "Error: Cannot open '" << filenames[i] << "'." << endl;
      goto error;
    }
  }

  if (!ss.decode(*out, files)) {
    cout << "Error: Stream did not decode properly." << endl;
    cout << "       Perhaps -k is not set properly, or the wrong "
      "keys" << endl;
    cout << "       are being combined." << endl;
    goto error;
  }

  for (int i=0; i<nfiles; i++)
    delete files[i];
  delete[] files;

  return;

  // Yes, this is a goto.  I feel no shame.

error:
  //out->close();			// let exit() close the file
  // unlink(output.c_str());
  return;
}

static void setInt(const char* str, int& val)
{
  val = atoi(str);
}

static void usage()
{
  cout << "Usage: " << endl;
  cout << "   tontine [options]" << endl << endl;

  cout << "General options:" << endl;
  cout << "   -d           Decode (Encoding is default)" << endl;
  cout << "   -S [string]  Encode string rather than file" << endl;
  cout << "   -t [binary|base64|words] Specify format of split files" << endl;
  cout << "   -s           Use steganography to conceal secrets" << endl;
  cout << "   -V           Show version and exit" << endl;
  cout << endl;

  cout << "Encoding options:" << endl;
  cout << "   -N [keys]    Create N keys" << endl;
  cout << "   -k [needed]  Require k keys to decode" << endl;
  cout << "   -p [prefix]  Use this prefix on keys" << endl;
  cout << endl;
  
  cout << "Decoding options:" << endl;
  cout << "   -o [file]    Output file (\"decoded\" is default)" << endl;
  cout << endl;

  exit(0);
}




